using System;
using System.Collections.Generic;
using System.IO;

using Smdn.Formats.Ini;
using Smdn.Windows.BaseServices;

namespace Smdn.Applications.SimpleHalt {
  public static class Config {
    public static readonly string DefaultConfigFileName = "SimpleHalt.ini";

    public static class HaltOptions {
      private static HaltOption @default;

      public static HaltOption Default {
        get
        {
          if (@default == null)
            @default = new HaltOption(Halt.HaltMethod.Shutdown, TimeSpan.FromHours(1.0));

          return @default;
        }
        set { @default = value; }
      }

      private static Dictionary<string, HaltOption> saved = new Dictionary<string, HaltOption>();

      public static Dictionary<string, HaltOption> Saved {
        get { return saved; }
      }

      internal static IEnumerable<HaltOption> EnumerateAllOptions()
      {
        yield return Default;

        foreach (var pair in saved) {
          yield return pair.Value;
        }
      }

      internal static void LoadFrom(IniDocument document)
      {
        foreach (var section in document) {
          if (!section.Name.StartsWith("HaltOption"))
            continue;

          var delim = section.Name.IndexOf(':');
          var name = (0 <= delim) ? section.Name.Substring(delim + 1) : null;

          var option = section.GetValue<HaltOption>("args", delegate(string val) {
            try {
              return HaltOption.CreateFromArgs(val);
            }
            catch {
              return null;
            }
          });

          if (option != null) {
            if (name == null)
              Default = option;
            else if (!saved.ContainsKey(name))
              saved.Add(name, option);
          }
        }
      }

      internal static void SaveTo(IniDocument document)
      {
        if (Default != null) {
          var section = document["HaltOption"];

          section["args"] = Default.GetJoinedCommandLineArgs();
        }

        foreach (var pair in saved) {
          var section = document[string.Format("HaltOption:{0}", pair.Key)];

          section["args"] = pair.Value.GetJoinedCommandLineArgs();
        }
      }
    }

    public static class ReportMail {
      private static readonly string SectionName = "ReportMail";

      private static List<string> addresses = new List<string>();

      public static List<string> Addresses {
        get { return addresses; }
      }

      internal static void LoadFrom(IniDocument document)
      {
        var section = document[SectionName];

        addresses.Clear();

        foreach (var entry in section) {
          if (entry.Key.StartsWith("MailAddress") && !string.IsNullOrEmpty(entry.Value) && !Addresses.Contains(entry.Value))
            addresses.Add(entry.Value);
        }
      }

      internal static void SaveTo(IniDocument document)
      {
        var section = document[SectionName];

        for (var i = 0; i < addresses.Count; i++) {
          section[string.Format("MailAddress{0}", i)] = addresses[i];
        }
      }

      internal static void MergeAddressesWithOptions(IEnumerable<HaltOption> options)
      {
        foreach (var option in options) {
          if (!string.IsNullOrEmpty(option.ReportMailToAddress) && !Addresses.Contains(option.ReportMailToAddress))
            addresses.Add(option.ReportMailToAddress);
        }
      }
    }

    public static void Load(string searchDirectory)
    {
      configFilePath = Path.Combine(searchDirectory, DefaultConfigFileName);

      if (!File.Exists(configFilePath)) {
        Console.Error.WriteLine("configuration file '{0}' not found", configFilePath);
        return;
      }

      try {
        var document = IniDocument.Load(configFilePath);

        HaltOptions.LoadFrom(document);
        ReportMail.LoadFrom(document);

        ReportMail.MergeAddressesWithOptions(HaltOptions.EnumerateAllOptions());
      }
      catch {
        // ignore any exceptions
      }
    }

    public static void Save()
    {
      try {
        var document = new IniDocument();

        HaltOptions.SaveTo(document);
        ReportMail.SaveTo(document);

        document.Save(configFilePath);
      }
      catch {
        // ignore any exceptions
      }
    }

    private static string configFilePath;
  }
}
