using System;
using System.ComponentModel;
using System.Threading;

using Smdn.Windows.BaseServices;
using Smdn.Windows.BaseServices.PlugAndPlay;

namespace Smdn.Applications.SimpleHalt {
  public class HaltOperationBackgroundWorker : BackgroundWorker {
    public HaltOperationBackgroundWorker()
      : base()
    {
    }

    public new void RunWorkerAsync()
    {
      throw new NotImplementedException();
    }

    public new void RunWorkerAsync(object argument)
    {
      if (argument is HaltOption)
        throw new ArgumentException("argument must be HaltOption");

      RunWorkerAsync(argument as HaltOption);
    }

    public void RunWorkerAsync(HaltOption option)
    {
      if (option == null)
        throw new ArgumentNullException("option");

      base.RunWorkerAsync(option);
    }

    protected override void OnDoWork(DoWorkEventArgs e)
    {
      var option = e.Argument as HaltOption;

      if (option == null)
        throw new InvalidOperationException("Argument must be HaltOption");

      e.Result = new HaltOperationResult(); // result for no-error or unexpected exception

      timeToHaltAtStarted = option.CalculateTimeToHalt();

      var keepActiveTarget = ThreadExecutionState.Target.None;

      if (option.KeepDisplayActive)
        // ディスプレイをアクティブ状態にしておく
        keepActiveTarget |= ThreadExecutionState.Target.Display;

      if (option.KeepSystemActive)
        // システムをアクティブ状態にしておく
        keepActiveTarget |= ThreadExecutionState.Target.System;

      try {
        if (keepActiveTarget != ThreadExecutionState.Target.None)
          ThreadExecutionState.KeepActive(keepActiveTarget);
      }
      catch (Win32Exception) {
        // ignore exception
      }

      ReportProgress(option, HaltOperationStatus.Ready, true);

      try {
        var doNotify = option.NotifyInterval.HasValue;
        var nextNotifyTime = doNotify ? DateTime.Now.Add(option.NotifyInterval.Value) : DateTime.MaxValue;

        for (;;) {
          // 中断が要求された場合
          if (CancellationPending) {
            ReportProgress(option, HaltOperationStatus.Aborted, true);

            e.Cancel = true;

            return;
          }

          // 指定時刻になった場合
          if (option.DoHaltAt <= DateTime.Now) {
            // シャットダウンを実行
            ReportProgress(option, HaltOperationStatus.Halting, true);

            try {
              if (option.Immediately)
                DoHaltImmediately(option);
              else
                DoHalt(option);

              ReportProgress(option, HaltOperationStatus.Halted, false);
            }
            catch (Exception ex) {
              ReportProgress(option, HaltOperationStatus.Failed, true);

              e.Result = new HaltOperationResult(ex); // overwrite result
            }

            return;
          }

          Thread.Sleep(250);

          // 一定時間毎に状態を通知する
          bool notifyProgress = false;

          if (doNotify && nextNotifyTime < DateTime.Now) {
            notifyProgress = true;
            nextNotifyTime = nextNotifyTime.Add(option.NotifyInterval.Value);
          }

          ReportProgress(option, HaltOperationStatus.Running, notifyProgress);
        } // for
      }
      finally {
        // アクティブにした状態を戻す
        try {
          if (keepActiveTarget != ThreadExecutionState.Target.None)
            ThreadExecutionState.RevertActivated(keepActiveTarget);
        }
        catch (Win32Exception) {
          // ignore exception
        }

        base.OnDoWork(e);
      }
    }

    private static void DoHaltImmediately(HaltOption option)
    {
      switch (option.Method) {
        case Halt.HaltMethod.LogOff:
        case Halt.HaltMethod.PowerOff:
        case Halt.HaltMethod.Reboot:
        case Halt.HaltMethod.Shutdown:
          Halt.ExitImmediately(Halt.ToExitMethod(option.Method));
          break;

        case Halt.HaltMethod.Suspend:
        case Halt.HaltMethod.Hibernate:
          Halt.Suspend(Halt.ToSuspendMethod(option.Method), true);
          break;

        case Halt.HaltMethod.Lock:
          Halt.Lock();
          break;
      }
    }

    private static void DoHalt(HaltOption option)
    {
      switch (option.Method) {
        case Halt.HaltMethod.LogOff:
        case Halt.HaltMethod.PowerOff:
        case Halt.HaltMethod.Reboot:
        case Halt.HaltMethod.Shutdown:
          Halt.Exit(Halt.ToExitMethod(option.Method), option.KillIfHung);
          break;

        case Halt.HaltMethod.Suspend:
        case Halt.HaltMethod.Hibernate:
          Halt.Suspend(Halt.ToSuspendMethod(option.Method), option.KillIfHung);
          break;

        case Halt.HaltMethod.Lock:
          Halt.Lock();
          break;
      } // switch
    }

    private void ReportProgress(HaltOption option, HaltOperationStatus status, bool notifyProgress)
    {
      var timeToHalt = option.CalculateTimeToHalt();
      var percentage = (int)(100.0 * timeToHalt.TotalSeconds / timeToHaltAtStarted.TotalSeconds);

      if (percentage < 0)
        percentage = 0;
      else if (100 < percentage)
        percentage = 100;

      var progress = new HaltOperationProgress(status,
                                               timeToHalt,
                                               percentage,
                                               GetStatusText(option, status));

      progress.NotifyProgress = notifyProgress;

      ReportProgress(progress.Percentage, progress);
    }

    private static string GetStatusText(HaltOption option, HaltOperationStatus status)
    {
      switch (status) {
        case HaltOperationStatus.Ready:
        case HaltOperationStatus.Running: {
          var timeToHalt = option.CalculateTimeToHalt();
          string timeToHaltString;

          if (0 < timeToHalt.Days)
            timeToHaltString = string.Format("{0}日{1:D2}時間{2:D2}分{3:D2}秒", timeToHalt.Days, timeToHalt.Hours, timeToHalt.Minutes, timeToHalt.Seconds);
          else if (0 < timeToHalt.Hours)
            timeToHaltString = string.Format("{0:D2}時間{1:D2}分{2:D2}秒", timeToHalt.Hours, timeToHalt.Minutes, timeToHalt.Seconds);
          else if (0 < timeToHalt.Minutes)
            timeToHaltString = string.Format("{0:D2}分{1:D2}秒", timeToHalt.Minutes, timeToHalt.Seconds);
          else
            timeToHaltString = string.Format("{0:D2}秒", timeToHalt.Seconds);

          if (status == HaltOperationStatus.Ready)
            return string.Format("{0}にシステムを{1}します{2}(あと{3})", option.DoHaltAt, option.MethodString, Environment.NewLine, timeToHaltString);
          else
            return string.Format("システムの{0}まで、あと{1}", option.MethodString, timeToHaltString);
        }

        case HaltOperationStatus.Halting: {
          string additionalMessage = null;

          switch (option.Method) {
            case Halt.HaltMethod.LogOff:
            case Halt.HaltMethod.PowerOff:
            case Halt.HaltMethod.Reboot:
            case Halt.HaltMethod.Shutdown:
              if (option.Immediately)
                additionalMessage = "即時停止します。　保存されていないデータが失われる場合があります。";
              else if (option.KillIfHung)
                additionalMessage = "応答しないアプリケーションがある場合は強制的に終了します。　保存されていないデータが失われる場合があります。";
              break;

            case Halt.HaltMethod.Suspend:
            case Halt.HaltMethod.Hibernate:
              if (!option.KillIfHung)
                additionalMessage = "起動中の他のアプリケーションによりキャンセルされる場合があります";
              break;
          }

          var ret = string.Format("システムを{0}しています。", option.MethodString);

          if (additionalMessage != null) {
            ret += Environment.NewLine;
            ret += additionalMessage;
          }

          return ret;
        }

        case HaltOperationStatus.Halted:
          return string.Format("システムを{0}しました", option.MethodString);

        case HaltOperationStatus.Aborted:
          return string.Format("システムの{0}を中断しました", option.MethodString);

        case HaltOperationStatus.Failed:
          return string.Format("エラーによりシステムの{0}に失敗しました", option.MethodString);

        default:
          return string.Empty;
      }
    }

    private TimeSpan timeToHaltAtStarted;
  }
}
