using System;
using System.IO;
using System.Collections.Generic;
using System.Globalization;
using System.Reflection;
using System.Text;

using Smdn.Windows.BaseServices;

namespace Smdn.Applications.SimpleHalt {
  public class HaltOption : ICloneable {
    [CommandLineArgument("method", ParameterName = "停止方法")]
    [CommandLineArgumentRequirement(true)]
    [CommandLineArgumentSummary("停止方法")]
    [CommandLineArgumentDescription(@"次のうちいずれかを指定してください。
shutdown     シャットダウン
logoff       ログオフ
reboot       再起動
poweroff     電源オフ
suspend      サスペンド
hibernate    休止状態
lock         ロック")]
    public Halt.HaltMethod Method {
      get; private set;
    }

    public bool IsSuspendLikeMethod {
      get
      {
        switch (Method) {
          case Halt.HaltMethod.Suspend:
          case Halt.HaltMethod.Hibernate:
          case Halt.HaltMethod.Lock:
            return true;
          default:
            return false;
        }
      }
    }

    public string MethodString {
      get { return GetHaltMethodString(Method); }
    }

    public DateTime DoHaltAt {
      get; private set;
    }

    [CommandLineArgument("now")]
    [CommandLineArgumentRequirement(true)]
    [CommandLineArgumentSummary("すぐに停止する")]
    public bool HaltNow {
      get; private set;
    }

    [CommandLineArgument("at", ParameterName = "日時")]
    [CommandLineArgumentRequirement(true)]
    [CommandLineArgumentSummary("指定した日時に停止する")]
    public DateTime? HaltAt {
      get; private set;
    }

    [CommandLineArgument("after", ParameterName = "時間")]
    [CommandLineArgumentRequirement(true)]
    [CommandLineArgumentSummary("指定した時間が経過した時点で停止する")]
    public TimeSpan? HaltAfter {
      get; private set;
    }

    [CommandLineArgument("i")]
    [CommandLineArgument("immediate")]
    [CommandLineArgument("f")]
    [CommandLineArgument("force")]
    [CommandLineArgumentSummary("すべてのアプリケーションを即座に終了する")]
    [CommandLineArgumentDescription(@"アプリケーションの終了処理を省略するため、
保存されていないデータが失われる場合があります。")]
    public bool Immediately {
      get; set;
    }

    [CommandLineArgument("k")]
    [CommandLineArgument("kill")]
    [CommandLineArgument("killifhung")]
    [CommandLineArgumentSummary("応答しないアプリケーションを強制的に終了する")]
    public bool KillIfHung {
      get; set;
    }

    [CommandLineArgument("ns")]
    [CommandLineArgument("nosleep")]
    [CommandLineArgumentSummary("停止するまではスリープ状態にしない")]
    [CommandLineArgumentDescription(@"システムの電源管理の設定でスリープ機能が有効になっている場合でも、
停止するまではコンピュータをスリープ状態に移行させないようにします。")]
    public bool KeepSystemActive {
      get; set;
    }

    [CommandLineArgument("nd")]
    [CommandLineArgument("nodisplayoff")]
    [CommandLineArgumentSummary("停止するまではディスプレイの電源を切らない")]
    [CommandLineArgumentDescription(@"システムの電源管理の設定でディスプレイの電源オフが有効になっている場合でも、
停止するまではディスプレイの電源をオフにしないようにします。")]
    public bool KeepDisplayActive {
      get; set;
    }

    [CommandLineArgument("w")]
    [CommandLineArgument("window")]
    [CommandLineArgumentSummary("コマンドラインからの起動時にウィンドウを表示する")]
    public bool ShowWindow {
      get; set;
    }

    [CommandLineArgument("nq")]
    [CommandLineArgument("noquit")]
    [CommandLineArgumentSummary("停止に成功した場合、SimpleHaltを終了しない")]
    [CommandLineArgumentDescription(@"停止が成功した場合でもSimpleHaltを起動したままにします。
停止方法がサスペンド、休止状態、ロックの場合のみ有効です。")]
    public bool DontQuit {
      get; set;
    }

    [CommandLineArgument("fq")]
    [CommandLineArgument("forcequit")]
    [CommandLineArgumentSummary("停止に失敗した場合、SimpleHaltを終了する")]
    [CommandLineArgumentDescription(@"停止中にエラーが発生した場合でもSimpleHaltを終了します。")]
    public bool ForceQuit {
      get; set;
    }

    [CommandLineArgument("n", ParameterName = "時間")]
    [CommandLineArgument("notify", ParameterName = "時間")]
    [CommandLineArgumentSummary("停止までの残り時間を通知する間隔を指定する")]
    [CommandLineArgumentDescription(@"指定された間隔で、停止までの残り時間を通知します。")]
    public TimeSpan? NotifyInterval {
      get; set;
    }

    [CommandLineArgument("r", ParameterName = "メールアドレス")]
    [CommandLineArgument("reportto", ParameterName = "メールアドレス")]
    [CommandLineArgumentSummary("実行結果を指定されたメールに送信する")]
    [CommandLineArgumentDescription(@"実行結果を指定されたメールに送信します。
送信に使用するSMTPサーバの設定は、SimpleHalt.exe.configに記述してください。")]
    public string ReportMailToAddress {
      get; set;
    }

    [CommandLineArgument("reportfrom", ParameterName = "メールアドレス")]
    [CommandLineArgumentSummary("メールの送信者アドレス")]
    [CommandLineArgumentDescription(@"実行結果をメールで送信する場合、指定されたアドレスを送信者アドレスとして使用します。
送信に使用するSMTPサーバの設定は、SimpleHalt.exe.configに記述してください。")]
    public string ReportMailFromAddress {
      get; set;
    }

    [CommandLineArgument("smtpssl")]
    [CommandLineArgumentSummary("SSLを使用してSTMPサーバに接続する")]
    [CommandLineArgumentDescription(@"実行結果をメールで送信する場合、SSLを使用してSTMPサーバに接続します。
送信に使用するSMTPサーバの設定は、SimpleHalt.exe.configに記述してください。")]
    public bool ReportMailSmtpEnableSsl {
      get; set;
    }

    public HaltOption(Halt.HaltMethod method)
      : this(method, DateTime.Now, false)
    {
      this.HaltNow = true;
      this.HaltAt = null;
      this.HaltAfter = null;
    }

    public HaltOption(Halt.HaltMethod method, DateTime haltAt)
      : this(method, haltAt, false)
    {
      if (TimeSpan.FromSeconds(2.0) < DateTime.Now.Subtract(haltAt))
        throw new ArgumentOutOfRangeException("haltAt", "must be representation of future date time");

      this.HaltNow = false;
      this.HaltAt = haltAt;
      this.HaltAfter = null;
    }

    public HaltOption(Halt.HaltMethod method, TimeSpan haltAfter)
      : this(method, DateTime.Now.Add(haltAfter), false)
    {
      if (haltAfter < TimeSpan.Zero)
        throw new ArgumentOutOfRangeException("haltAfter", "must be zero or positive number");

      this.HaltNow = false;
      this.HaltAt = null;
      this.HaltAfter = haltAfter;
    }

    private HaltOption(Halt.HaltMethod method, DateTime doHaltAt, bool notUsed)
    {
      this.Method = method;
      this.DoHaltAt = doHaltAt;

      this.Immediately = false;
      this.KillIfHung = false;
      this.KeepSystemActive = false;
      this.KeepDisplayActive = false;
      this.ShowWindow = false;
      this.DontQuit = false;
      this.ForceQuit = false;
      this.NotifyInterval = null;
      this.ReportMailToAddress = null;
      this.ReportMailFromAddress = null;
      this.ReportMailSmtpEnableSsl = false;
    }

    public HaltOption Clone()
    {
      return (HaltOption)MemberwiseClone();
    }

    object ICloneable.Clone()
    {
      return Clone();
    }

    public TimeSpan CalculateTimeToHalt()
    {
      return DoHaltAt.Subtract(DateTime.Now);
    }

    public string GetJoinedCommandLineArgs()
    {
      return string.Join(" ", GetCommandLineArgs());
    }

    public string[] GetCommandLineArgs()
    {
      var args = new List<string>();

      foreach (var propertyInfo in GetType().GetProperties(BindingFlags.Public | BindingFlags.Instance)) {
        var argAttr = propertyInfo.GetSingleCustomAttribute<CommandLineArgumentAttribute>();

        if (argAttr == null)
          continue;

        var arg = argAttr.Argument;

        if (propertyInfo.PropertyType == typeof(bool) && (bool)propertyInfo.GetValue(this, null) == true)
          args.Add(string.Format("/{0}", arg));
        else if (propertyInfo.PropertyType == typeof(string) && (string)propertyInfo.GetValue(this, null) != null)
          args.Add(string.Format("/{0}:{1}", arg, (string)propertyInfo.GetValue(this, null)));
        else if (propertyInfo.PropertyType == typeof(TimeSpan))
          args.Add(string.Format("/{0}:{1}", arg, (TimeSpan)propertyInfo.GetValue(this, null)));
        else if (propertyInfo.PropertyType == typeof(TimeSpan?) && (TimeSpan?)propertyInfo.GetValue(this, null) != null)
          args.Add(string.Format("/{0}:{1}", arg, (TimeSpan)propertyInfo.GetValue(this, null)));
        else if (propertyInfo.PropertyType == typeof(DateTime?) && (DateTime?)propertyInfo.GetValue(this, null) != null)
          args.Add(string.Format("/{0}:{1:yyyy/M/d;H:m:s}", arg, (DateTime)propertyInfo.GetValue(this, null)));
        else if (propertyInfo.PropertyType == typeof(Halt.HaltMethod))
          args.Add(string.Format("/{0}:{1}", arg, (Halt.HaltMethod)propertyInfo.GetValue(this, null)));
      }

      return args.ToArray();
    }

    public string GetShortDescriptionString()
    {
      if (HaltAt.HasValue) {
        var haltAt = HaltAt.Value;

        if (haltAt.Date == DateTime.Now.Date)
          // today
          return string.Format("{0}に{1}", haltAt.ToString("T", CultureInfo.CurrentUICulture), GetHaltMethodString(Method));
        else
          return string.Format("{0}に{1}", haltAt.ToString("F", CultureInfo.CurrentUICulture), GetHaltMethodString(Method));
      }
      else if (HaltAfter.HasValue) {
        var haltAfter = HaltAfter.Value;
        var sb = new StringBuilder();

        if (0 < haltAfter.Days)
          sb.AppendFormat("{0}日", haltAfter.Days);
        if (0 < haltAfter.Hours)
          sb.AppendFormat("{0}時間", haltAfter.Hours);
        if (0 < haltAfter.Minutes)
          sb.AppendFormat("{0}分", haltAfter.Minutes);
        if (0 < haltAfter.Seconds)
          sb.AppendFormat("{0}秒", haltAfter.Seconds);

        sb.Append("後に");
        sb.Append(GetHaltMethodString(Method));

        return sb.ToString();
      }
      else {
        return GetHaltMethodString(Method);
      }
    }

    internal static HaltOption CreateFromArgs(string args)
    {
      return CreateFromArgs(args.Split(new[] {' '}, StringSplitOptions.RemoveEmptyEntries));
    }

    internal static HaltOption CreateFromArgs(string[] args)
    {
      // mandatory args
      Halt.HaltMethod? method = null;
      DateTime? haltAt = null;
      TimeSpan? haltAfter = null;
      var haltNow = false;
      var unparsedArgs = new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase);

      foreach (var a in args) {
        string arg;

        if ((a.StartsWith("/") || a.StartsWith("-")))
          arg = a.Substring(1);
        else if (a.StartsWith("--"))
          arg = a.Substring(2);
        else
          // invalid argument format
          continue;

        var delim = arg.IndexOf(':');
        string param;

        if (0 <= delim) {
          param = arg.Substring(delim + 1);
          arg = arg.Substring(0, delim);
        }
        else {
          param = string.Empty;
        }

        switch (arg.ToLowerInvariant()) {
          case "at":    haltAt    = ConvertOrThrowDateTime(param); break;
          case "after": haltAfter = ConvertOrThrowTimeSpan(param); break;
          case "now":   haltNow = true; break;

          case "method":
            ConvertOrThrowHaltMethod(ref method, param, false);
            break;

          default:
            if (!ConvertOrThrowHaltMethod(ref method, arg, true))
              unparsedArgs.Add(arg, param);
            break;
        }
      } // foreach

      if (method == null)
        throw new ArgumentException("停止方法を指定してください。");

      if (haltAt == null && haltAfter == null && !haltNow)
        throw new ArgumentException("停止時間を指定してください");

      HaltOption option = null;

      try {
        if (haltAt.HasValue)
          // this will throw ArgumentException
          option = new HaltOption(method.Value, haltAt.Value);
        else if (haltAfter.HasValue)
          // this will throw ArgumentException
          option = new HaltOption(method.Value, haltAfter.Value);
        else // if (haltNow)
          option = new HaltOption(method.Value);
      }
      catch {
        throw new ArgumentException("不正な停止時間が指定されました");
      }

      // optional args
      foreach (var propertyInfo in option.GetType().GetProperties(BindingFlags.Public | BindingFlags.Instance | BindingFlags.SetProperty)) {
        var found = false;
        string arg = null;
        string param = null;

        foreach (CommandLineArgumentAttribute argAttr in propertyInfo.GetCustomAttributes(typeof(CommandLineArgumentAttribute), true)) {
          if (unparsedArgs.ContainsKey(argAttr.Argument)) {
            arg = argAttr.Argument;
            param = unparsedArgs[arg];
            found = true;
            break;
          }
        }

        if (!found)
          continue;

        if (propertyInfo.PropertyType == typeof(bool))
          propertyInfo.SetValue(option, true, null);
        else if (propertyInfo.PropertyType == typeof(string))
          propertyInfo.SetValue(option, param, null);
        else if (propertyInfo.PropertyType == typeof(TimeSpan?))
          propertyInfo.SetValue(option, ConvertOrThrowTimeSpan(param), null);
        else
          continue;

        // parsed
        unparsedArgs.Remove(arg);
      }

      foreach (var pair in unparsedArgs) {
        Console.Error.WriteLine("ignore: {0}={1}", pair.Key, pair.Value);
      }

      return option;
    }

    private static bool ConvertOrThrowHaltMethod(ref Halt.HaltMethod? method, string val, bool ignoreInvalid)
    {
      Halt.HaltMethod parsedMethod;
      var valid = EnumUtils.TryParse(val, true, out parsedMethod);

      if (valid) {
        if (method == null)
          method = parsedMethod;
        else
          throw new ArgumentException("設定出来る停止方法は1つだけです。");
      }

      if (ignoreInvalid)
        return valid;
      else if (!valid)
        throw new ArgumentException("停止方法が不正です。");
      else
        return true;
    }

    private static readonly string[] dateTimeFormats = new[] {
      "yyyy/M/d;H:m:s",
      "yyyy/M/d;H:m",
      "yyyy/M/d",
      "yy/M/d;H:m:s",
      "yy/M/d;H:m",
      "yy/M/d",
      "M/d;H:m:s",
      "M/d;H:m",
      "M/d",
      "H:m:s",
      "H:m",
    };

    private static DateTime ConvertOrThrowDateTime(string val)
    {
      try {
        return DateTime.ParseExact(val, dateTimeFormats, CultureInfo.CurrentCulture, DateTimeStyles.AssumeLocal);
      }
      catch (FormatException) {
        throw new ArgumentException("日時の形式が不正です");
      }
    }

    private static TimeSpan ConvertOrThrowTimeSpan(string val)
    {
      try {
        var parsed = TimeSpan.Parse(val);

        if (parsed < TimeSpan.Zero)
          throw new ArgumentException("時刻が負の値を表しています");
        else
          return parsed;
      }
      catch (FormatException) {
        throw new ArgumentException("時刻の形式が不正です");
      }
    }

    internal static string GetHaltMethodString(Halt.HaltMethod method)
    {
      switch (method) {
        case Halt.HaltMethod.LogOff:    return "ログオフ";
        case Halt.HaltMethod.PowerOff:  return "電源オフ";
        case Halt.HaltMethod.Reboot:    return "再起動";
        case Halt.HaltMethod.Shutdown:  return "シャットダウン";
        case Halt.HaltMethod.Suspend:   return "サスペンド";
        case Halt.HaltMethod.Hibernate: return "休止状態";
        case Halt.HaltMethod.Lock:      return "ロック";
        default: throw new ArgumentException("invalid method", "method");
      }
    }

    internal static void WriteUsage(TextWriter writer)
    {
      var mandatoryArgProperties = new List<PropertyInfo>();
      var optionalArgProperties = new List<PropertyInfo>();

      foreach (var propertyInfo in typeof(HaltOption).GetProperties(BindingFlags.Public | BindingFlags.Instance)) {
        var requirementAttr = propertyInfo.GetSingleCustomAttribute<CommandLineArgumentRequirementAttribute>();

        if (requirementAttr != null && requirementAttr.Required)
          mandatoryArgProperties.Add(propertyInfo);
        else
          optionalArgProperties.Add(propertyInfo);
      }

      // mandatory args
      writer.WriteLine("必須の引数:");

      foreach (var argProperty in mandatoryArgProperties) {
        if (WriteUsage(writer, argProperty))
          writer.WriteLine();
      }

      writer.WriteLine();

      // optional args
      writer.WriteLine("省略可能な引数:");

      foreach (var argProperty in optionalArgProperties) {
        if (WriteUsage(writer, argProperty))
          writer.WriteLine();
      }

      // formats
      writer.WriteLine("日時と時間の形式:");
      writer.WriteLine("  日時は次の形式で指定可能です");
      writer.WriteLine("  ※時間は0〜23の間の値で指定してください");

      foreach (var format in dateTimeFormats) {
        var f = format;

        f = f.Replace("yyyy", "年(西暦4桁)");
        f = f.Replace("yy", "年(西暦下2桁)");
        f = f.Replace("M", "月");
        f = f.Replace("d", "日");
        f = f.Replace("H", "時");
        f = f.Replace("m", "分");
        f = f.Replace("s", "秒");

        writer.WriteLine("    ・{0}", f);
      }

      writer.WriteLine();

      writer.WriteLine("  時間は次の形式で指定可能です。");
      writer.WriteLine("  ※時間は0〜23、分・秒は0〜59の間の値で指定してください");
      writer.WriteLine("    ・日.時間:分:秒");
      writer.WriteLine("    ・日.時間:分");
      writer.WriteLine("    ・日");
      writer.WriteLine("    ・時間:分:秒");
      writer.WriteLine("    ・時間:分");
      writer.WriteLine();

      // examples
      writer.WriteLine("オプションの例:");
      writer.WriteLine("  /shutdown /now");
      writer.WriteLine("    今すぐシャットダウンする");
      writer.WriteLine();
      writer.WriteLine("  /lock /at:09/12/7;6:0 /noquit /nosleep");
      writer.WriteLine("    2009年12月7日 午前6時にロックする");
      writer.WriteLine("    ロックした後もSimpleHaltを起動したままにする");
      writer.WriteLine("    ロックするまでの間、スリープさせない");
      writer.WriteLine();
      writer.WriteLine("  /reboot /after:3.12:00 /notify:2:0");
      writer.WriteLine("    3日と12時間後に再起動する");
      writer.WriteLine("    再起動するまでの間、2時間おきに通知する");
      writer.WriteLine();
      writer.WriteLine("  /poweroff /at:3:00 /force /reportto:alert@mail.example.com");
      writer.WriteLine("    今日の午前3時に電源をオフにする");
      writer.WriteLine("    応答のないアプリケーションがある場合でも強制的に停止する");
      writer.WriteLine("    停止する際に、結果をalert@mail.example.com宛にメールで送信する");
      writer.WriteLine();
    }

    private static bool WriteUsage(TextWriter writer, PropertyInfo argProperty)
    {
      const string indent = "  ";
      const int descriptionIndentLevel = 34;

      var summaryAttr = argProperty.GetSingleCustomAttribute<CommandLineArgumentSummaryAttribute>();
      var summary = (summaryAttr == null) ? null : summaryAttr.Summary;

      var descriptionAttr = argProperty.GetSingleCustomAttribute<CommandLineArgumentDescriptionAttribute>();
      var description = (descriptionAttr == null) ? null : descriptionAttr.Description;

      var args = string.Empty;
      var isFirst = true;

      // XXX
      if (argProperty.PropertyType == typeof(Halt.HaltMethod)) {
        args += "/[停止方法]";
        isFirst = false;
      }

      foreach (CommandLineArgumentAttribute argAttr in argProperty.GetCustomAttributes(typeof(CommandLineArgumentAttribute), true)) {
        if (isFirst)
          isFirst = false;
        else
          args += ", ";

        args += string.Format("/{0}", argAttr.Argument);

        if (argAttr.ParameterName != null)
          args += string.Format(":[{0}]", argAttr.ParameterName);
      }

      if (args.Length == 0)
        // CommandLineArgumentAttribute is not applied
        return false;

      writer.Write(indent);

      if (summary == null) {
        writer.WriteLine(args);
      }
      else {
        var padding = new string(' ', Math.Max(1, descriptionIndentLevel - Encoding.GetEncoding("shift_jis").GetByteCount(args))); // XXX

        writer.WriteLine("{0}{1}{2}", args, padding, summary);
      }

      if (description == null)
        return true;

      using (var reader = new StringReader(description)) {
        for (;;) {
          var line = reader.ReadLine();

          if (line == null)
            break;

          writer.WriteLine("{0}{1}{2}", indent, new string(' ', descriptionIndentLevel), line);
        }
      }

      return true;
    }
  }
}
