using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows.Forms;

using Smdn.IO;
using Smdn.Windows.BaseServices;
using Smdn.Windows.Forms;

namespace Smdn.Applications.SimpleHalt {
  public partial class MainDialog : Form {
    private class HaltMethodDropDownList : FormattingDropDownList<Halt.HaltMethod> {
      public HaltMethodDropDownList()
      {
        StringFormatter = delegate(Halt.HaltMethod method) {
          return HaltOption.GetHaltMethodString(method);
        };

        Items.Add(Halt.HaltMethod.PowerOff);
        Items.Add(Halt.HaltMethod.Shutdown);
        Items.Add(Halt.HaltMethod.Reboot);
        Items.Add(Halt.HaltMethod.LogOff);
        Items.Add(Halt.HaltMethod.Suspend);
        Items.Add(Halt.HaltMethod.Hibernate);
        Items.Add(Halt.HaltMethod.Lock);
      }
    }

    private SimpleHaltApplicationContext appContext;

    internal SimpleHaltApplicationContext AppContext {
      get { return appContext; }
      set
      {
        appContext = value;
        SwitchView(DetailView);
      }
    }

    private bool detailView = false;

    private bool DetailView {
      get { return detailView; }
      set
      {
        if (detailView != value)
          SwitchView(value);
        detailView = value;
      }
    }
    
    public MainDialog()
    {
      InitializeComponent();

      Initialize();
    }

    private void Initialize()
    {
      dropDownListHaltMethod.SelectedIndex = 0;

      timeSpanUpDownHaltAfter.Minimum = TimeSpan.Zero;
      timeSpanUpDownHaltAfter.Maximum = TimeSpan.FromTicks(long.MaxValue);
      timeSpanUpDownHaltAfter.Increment = TimeSpan.FromMinutes(1.0);
      timeSpanUpDownHaltAfter.TimeSpanToStringConverter = delegate(TimeSpan val) {
        if (0 < val.Days)
          return string.Format("{0}日 {1}時間 {2}分 {3}秒", (int)val.TotalDays, val.Hours, val.Minutes, val.Seconds);
        else if (0 < val.Hours)
          return string.Format("{0}時間 {1}分 {2}秒", (int)val.TotalHours, val.Minutes, val.Seconds);
        else if (0 < val.Minutes)
          return string.Format("{0}分 {1}秒", (int)val.TotalMinutes, val.Seconds);
        else
          return string.Format("{0}秒", (int)val.TotalSeconds);
      };
      timeSpanUpDownHaltAfter.StringToTimeSpanConverter = StringToTimeSpanConverter;

      timeSpanUpDownNotifyInterval.Minimum = TimeSpan.Zero;
      timeSpanUpDownNotifyInterval.Maximum = TimeSpan.FromTicks(long.MaxValue);
      timeSpanUpDownNotifyInterval.Increment = TimeSpan.FromMinutes(1.0);
      timeSpanUpDownNotifyInterval.TimeSpanToStringConverter = delegate(TimeSpan val) {
        if (0 < val.Hours)
          return string.Format("{0}時間 {1}分", (int)val.TotalHours, val.Minutes);
        else
          return string.Format("{0}分", (int)val.TotalMinutes);
      };
      timeSpanUpDownNotifyInterval.StringToTimeSpanConverter = StringToTimeSpanConverter;

      checkBoxImmediately.CheckedChanged += new EventHandler(checkBoxImmediately_CheckedChanged);

      checkBoxReportByMail.CheckedChanged += delegate {
        comboBoxReportMailToAddresses.Enabled = 
        checkBoxReportMailSmtpEnableSsl.Enabled = 
        textBoxReportMailFromAddress.Enabled =
        buttonSendTestReport.Enabled = checkBoxReportByMail.Checked;
      };

      comboBoxReportMailToAddresses.Items.Clear();
      comboBoxReportMailToAddresses.Items.AddRange(Config.ReportMail.Addresses.ToArray());

      timeSpanUpDownNotifyInterval.ValueChanged += delegate { checkBoxNotify.Checked = true; };

      dateTimePickerHaltAt.GotFocus += delegate { radioButtonHaltAt.Checked = true; };
      dateTimePickerHaltAt.ValueChanged += delegate {
        if (radioButtonHaltAt.Checked) {
          var haltAfter = dateTimePickerHaltAt.Value.Subtract(DateTime.Now);

          if (haltAfter < TimeSpan.Zero)
            haltAfter = TimeSpan.Zero;

          timeSpanUpDownHaltAfter.Value = haltAfter;
        }
      };

      timeSpanUpDownHaltAfter.GotFocus += delegate { radioButtonHaltAfter.Checked = true; };
      timeSpanUpDownHaltAfter.ValueChanged += delegate {
        if (radioButtonHaltAfter.Checked)
          dateTimePickerHaltAt.Value = DateTime.Now.Add(timeSpanUpDownHaltAfter.Value);
      };

      buttonSwitchView.Click += delegate { DetailView = !DetailView; };

      progressBarCountDown.Minimum = 0;
      progressBarCountDown.Maximum = 100;

      toolStripMenuItemFile_CreateShortcutWithCurrentOption_SaveAs.Click += delegate { CreateShortcut(); };
      toolStripMenuItemFile_CreateShortcutWithCurrentOption_OnQuickLaunch.Click += delegate { CreateShortcutOnQuickLaunch(); };
      toolStripMenuItemFile_Quit.Click += delegate { Close(); };

      toolStripMenuItemHelp_Usage.Click += delegate {
        using (var dialog = new UsageDialog()) {
          dialog.ShowDialog(this);
        }
      };

      detailView = false;
    }

    private Regex timeSpanRegex = new Regex(@"((?<days>\d+)\s*日)?\s*((?<hours>\d+)\s*時間)?\s*((?<minutes>\d+)\s*分)?\s*((?<seconds>\d+)\s*秒)?", RegexOptions.Singleline);

    private TimeSpan StringToTimeSpanConverter(string str)
    {
      int val;

      if (int.TryParse(str, out val))
        return TimeSpan.FromMinutes(val);

      var match = timeSpanRegex.Match(str);

      if (!match.Success)
        throw new FormatException();

      var days    = match.Groups["days"].Success    ? int.Parse(match.Groups["days"].Value)    : 0;
      var hours   = match.Groups["hours"].Success   ? int.Parse(match.Groups["hours"].Value)   : 0;
      var minutes = match.Groups["minutes"].Success ? int.Parse(match.Groups["minutes"].Value) : 0;
      var seconds = match.Groups["seconds"].Success ? int.Parse(match.Groups["seconds"].Value) : 0;

      return TimeSpan.FromSeconds(days * 86400.0 + hours * 3600.0 + minutes * 60.0 + seconds);
    }

    internal void ApplyOption(HaltOption option)
    {
      option = option ?? Config.HaltOptions.Default;

      dropDownListHaltMethod.SelectedItem = option.Method;

      checkBoxKillIfHung.Checked = option.KillIfHung;
      checkBoxImmediately.Checked = option.Immediately;

      checkBoxKeepDisplayActive.Checked = option.KeepDisplayActive;
      checkBoxKeepSystemActive.Checked = option.KeepSystemActive;

      if (!option.NotifyInterval.HasValue) {
        timeSpanUpDownNotifyInterval.Value = TimeSpan.FromMinutes(30.0);
        checkBoxNotify.Checked = false;
      }
      else {
        timeSpanUpDownNotifyInterval.Value = option.NotifyInterval.Value;
        checkBoxNotify.Checked = true;
      }

      checkBoxDontQuit.Checked = option.DontQuit;
      checkBoxForceQuit.Checked = option.ForceQuit;

      checkBoxReportByMail.CheckState = CheckState.Indeterminate; // XXX: to call CheckedChanged
      checkBoxReportByMail.CheckState = CheckState.Checked;
      checkBoxReportByMail.Checked = !string.IsNullOrEmpty(option.ReportMailToAddress);
      comboBoxReportMailToAddresses.Text = option.ReportMailToAddress;
      textBoxReportMailFromAddress.Text = option.ReportMailFromAddress;
      checkBoxReportMailSmtpEnableSsl.Checked = option.ReportMailSmtpEnableSsl;

      if (option.HaltAt.HasValue) {
        radioButtonHaltAt.Checked = true;

        var haltAt = option.HaltAt.Value;

        if (haltAt < dateTimePickerHaltAt.MinDate)
          dateTimePickerHaltAt.Value = dateTimePickerHaltAt.MinDate;
        else if (dateTimePickerHaltAt.MaxDate < haltAt)
          dateTimePickerHaltAt.Value = dateTimePickerHaltAt.MaxDate;
        else
          dateTimePickerHaltAt.Value = haltAt;
      }
      else if (option.HaltAfter.HasValue) {
        radioButtonHaltAfter.Checked = true;

        var haltAfter = option.HaltAfter.Value;

        if (haltAfter < timeSpanUpDownHaltAfter.Minimum)
          timeSpanUpDownHaltAfter.Value = timeSpanUpDownHaltAfter.Minimum;
        else if (timeSpanUpDownHaltAfter.Maximum < haltAfter)
          timeSpanUpDownHaltAfter.Value = timeSpanUpDownHaltAfter.Maximum;
        else
          timeSpanUpDownHaltAfter.Value = haltAfter;
      }
      else if (option.HaltNow) {
        radioButtonHaltNow.Checked = true;
      }
    }

    protected override void OnClosing(CancelEventArgs e)
    {
      Config.HaltOptions.Default = GetCurrentOption(false) ?? Config.HaltOptions.Default;

      base.OnClosing(e);
    }

    private void SwitchView(bool detail)
    {
      panelCountDown.Left = panelSimple.Left;

      if (detail) {
        buttonSwitchView.Text = "<<簡易設定";

        panelDetail.Visible = true;

        panelCountDown.Top = panelDetail.Bottom + 4;
        panelCountDown.Width = panelDetail.Right - panelSimple.Left;
      }
      else {
        buttonSwitchView.Text = "詳細設定>>";

        panelDetail.Visible = false;

        panelCountDown.Top = panelSimple.Bottom + 4;
        panelCountDown.Width = panelSimple.Width;
      }

      if (AppContext.IsHaltOperationRunning)
        SetAbortableAllignment();
      else
        SetHaltableAllignment();
    }

    internal void SetHaltableAllignment()
    {
      if (InvokeRequired) {
        Invoke((MethodInvoker)SetHaltableAllignment);
        return;
      }

      panelCountDown.Visible = false;

      buttonHaltOrAbort.Text = "実行";
      buttonHaltOrAbort.Top = panelCountDown.Top;

      groupBoxHaltMethod.Enabled =
      groupBoxHaltOption.Enabled =
      groupBoxNotificationOption.Enabled = true;

      progressBarCountDown.Value = 0;

      SetUpdatedAlignment();
    }

    internal void SetAbortableAllignment()
    {
      if (InvokeRequired) {
        Invoke((MethodInvoker)SetHaltableAllignment);
        return;
      }

      panelCountDown.Visible = true;

      buttonHaltOrAbort.Text = "中止";
      buttonHaltOrAbort.Top = panelCountDown.Bottom + 4;

      groupBoxHaltMethod.Enabled =
      groupBoxHaltOption.Enabled =
      groupBoxNotificationOption.Enabled = false;

      SetUpdatedAlignment();
    }

    private void SetUpdatedAlignment()
    {
      this.ClientSize = new Size(panelCountDown.Right + panelCountDown.Left, buttonHaltOrAbort.Bottom + 10);

      buttonHaltOrAbort.Left = (ClientSize.Width - buttonHaltOrAbort.Width) / 2;
    }

    // 実行/中断する
    private void buttonHaltOrAbort_Click(object sender, EventArgs e)
    {
      if (AppContext.IsHaltOperationRunning)
        DoAbort();
      else
        DoHalt();
    }

    private void DoHalt()
    {
      if (AppContext.IsHaltOperationRunning)
        // TODO: MessageBox.Show
        return;

      // 実行する
      var option = GetCurrentOption(true);

      if (option == null)
        return; // canceled

      Config.HaltOptions.Default = option;

      AppContext.StartHaltOperation(option);
    }

    private void DoAbort()
    {
      // 中断する
      AppContext.AbortHaltOperation();
    }

    private void checkBoxImmediately_CheckedChanged(object sender, EventArgs e)
    {
      if (checkBoxImmediately.Checked) {
        checkBoxKillIfHung.Checked = true;
        checkBoxKillIfHung.Enabled = false;
      }
      else {
        checkBoxKillIfHung.Enabled = true;
      }
    }

    internal void DisplayProgress(HaltOperationProgress progress)
    {
      progressBarCountDown.Value = progress.Percentage;
      labelCountDown.Text = progress.StatusText;
    }

    private HaltOption GetCurrentOption(bool allowShowingMessageBox)
    {
      // 停止方法
      if (dropDownListHaltMethod.SelectedIndex == ListBox.NoMatches) {
        if (allowShowingMessageBox)
          MessageBox.Show(this, "停止方法を選択してください。", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Information);
        return null;
      }

      // 実行時間
      HaltOption option;

      if (radioButtonHaltAt.Checked) {
        if (dateTimePickerHaltAt.Value < DateTime.Now) {
          if (allowShowingMessageBox)
            MessageBox.Show(this, "現在時刻より未来の時刻を指定してください。", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Information);
          return null;
        }

        option = new HaltOption(dropDownListHaltMethod.SelectedItem, dateTimePickerHaltAt.Value);
      }
      else if (radioButtonHaltAfter.Checked) {
        option = new HaltOption(dropDownListHaltMethod.SelectedItem, timeSpanUpDownHaltAfter.Value);
      }
      else if (radioButtonHaltNow.Checked) {
        option = new HaltOption(dropDownListHaltMethod.SelectedItem);
      }
      else {
        if (allowShowingMessageBox)
          MessageBox.Show(this, "停止時刻を選択してください。", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Information);
        return null;
      }

      // 即座に停止する
      option.Immediately = checkBoxImmediately.Checked;

      // 応答しないアプリケーションを終了する
      option.KillIfHung = checkBoxKillIfHung.Checked;

      // 起動中はディスプレイの電源を切らない
      option.KeepDisplayActive = checkBoxKeepDisplayActive.Checked;

      // 起動中はスクリーンセーバーを起動しない
      option.KeepSystemActive = checkBoxKeepSystemActive.Checked;

      // 通知間隔
      option.NotifyInterval = checkBoxNotify.Checked ? (TimeSpan?)timeSpanUpDownNotifyInterval.Value : null;

      option.DontQuit = checkBoxDontQuit.Checked;
      option.ForceQuit = checkBoxForceQuit.Checked;

      if (checkBoxReportByMail.Checked && !string.IsNullOrEmpty(comboBoxReportMailToAddresses.Text)) {
        option.ReportMailToAddress = comboBoxReportMailToAddresses.Text;
        option.ReportMailFromAddress = textBoxReportMailFromAddress.Text;
        option.ReportMailSmtpEnableSsl = checkBoxReportMailSmtpEnableSsl.Checked;
      }

      return option;
    }

    private void CreateShortcut()
    {
      var option = GetCurrentOption(true);

      if (option == null)
        return; // canceled

      using (var dialog = new SaveFileDialog()) {
        dialog.DefaultExt = Smdn.Windows.UserInterfaces.Shells.ShellLink.Extension;
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {"ショートカットファイル", "*" + Smdn.Windows.UserInterfaces.Shells.ShellLink.Extension},
          new[] {"すべてのファイル", "*.*"},
        });
        dialog.FilterIndex = 0;
        dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
        dialog.SupportMultiDottedExtensions = true;
        dialog.ValidateNames = true;
        dialog.FileName = PathUtils.ReplaceInvalidFileNameChars(option.GetShortDescriptionString(), ReplaceInvalidFileNameCharsEvaluator);

        if (DialogResult.OK == dialog.ShowDialog(this))
          CreateShortcut(option, dialog.FileName);
      }
    }

    private void CreateShortcutOnQuickLaunch()
    {
      var option = GetCurrentOption(true);

      if (option == null)
        return; // canceled

      string quickLaunchPath;

      if (Environment.OSVersion.Platform == PlatformID.Win32NT && 6 <= Environment.OSVersion.Version.Major) {
        if (2 <= Environment.OSVersion.Version.Minor)
          // 7 or over
          quickLaunchPath = @"Microsoft\Internet Explorer\Quick Launch\User Pinned\TaskBar";
        else
          // Vista, 2008
          quickLaunchPath = @"Microsoft\Internet Explorer\Quick Launch";
      }
      else {
        // XP or else
        quickLaunchPath = @"Microsoft\Internet Explorer\Quick Launch";
      }

      var directory = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), quickLaunchPath);
      var fileName = PathUtils.ReplaceInvalidFileNameChars(string.Format("{0} {1}.lnk", Application.ProductName, option.GetShortDescriptionString()), ReplaceInvalidFileNameCharsEvaluator);

      CreateShortcut(option, Path.Combine(directory, fileName));
    }

    private void CreateShortcut(HaltOption option, string path)
    {
      try {
        SimpleHalt.CreateShortcut(option, path);

        MessageBox.Show(this, "ショートカットを作成しました。", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Information);
      }
      catch (Exception ex) {
        MessageBox.Show(this, string.Format("ショートカットを作成できませんでした。{0}{1}", Environment.NewLine, ex), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Warning);
      }
    }

    private static string ReplaceInvalidFileNameCharsEvaluator(char ch, string str, int index)
    {
      if (ch == ':')
        return "：";
      else
        return " ";
    }

    private void buttonSendTestReport_Click(object sender, EventArgs e)
    {
      var toAddress = comboBoxReportMailToAddresses.Text;
      var fromAddress = textBoxReportMailFromAddress.Text;

      if (string.IsNullOrEmpty(toAddress)) {
        MessageBox.Show(this, "送信先のメールアドレスを入力してください。", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
        return;
      }

      var confirmationMessage = string.Format("{0}宛てに確認用のテストメールを送信します。", toAddress);

      if (string.IsNullOrEmpty(fromAddress))
        confirmationMessage += string.Format("{0}(デフォルトの送信者アドレスとして{1}を使います。)", Environment.NewLine, ReportMailSender.GetDefaultFromAddress());

      var result = MessageBox.Show(this, confirmationMessage, Application.ProductName, MessageBoxButtons.OKCancel, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);

      if (result != DialogResult.OK)
        return;

      try {
        Cursor = Cursors.WaitCursor;

        Application.DoEvents();

        try {
          ReportMailSender.SendTest(checkBoxReportMailSmtpEnableSsl.Checked, toAddress, fromAddress);

          MessageBox.Show(this, "送信しました。", Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Information);
        }
        catch (Exception ex) {
          MessageBox.Show(this, string.Format("エラーにより送信できませんでした。{0}{1}", Environment.NewLine, ex), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
        }
      }
      finally {
        Cursor = Cursors.Default;
      }
    }
  }
}