using System;
using System.ComponentModel;
using System.Windows.Forms;

using Smdn.Windows.BaseServices;
using Smdn.Windows.Forms;

namespace Smdn.Applications.SimpleHalt {
  public class SimpleHaltApplicationContext : NotifyIconApplicationContext {
    public bool IsHaltOperationRunning {
      get { return haltOperationWorker.IsBusy; }
    }

    private MainDialog MainDialog {
      get { return MainForm == null ? null : MainForm.IsDisposed ? null : MainForm as MainDialog; }
    }

    public SimpleHaltApplicationContext(HaltOption option)
    {
      components = new Container();

      toolStripMenuItemStatusSeparator = new ToolStripSeparator();
      toolStripMenuItemStatusSeparator.Visible = false;

      toolStripMenuItemStatus = new ToolStripMenuItem(Application.ProductName);
      toolStripMenuItemStatus.Visible = false;

      contextMenu = new ContextMenuStrip(components);
      contextMenu.Items.AddRange(new ToolStripItem[] {
        new ToolStripMenuItem("ウィンドウを表示(&S)", null, delegate { ShowDialog(); }),
        toolStripMenuItemStatusSeparator,
        toolStripMenuItemStatus,
        //new ToolStripMenuItem("バージョン情報(&V)...", null, delegate { ShowVersionInfoDialog(); }),
        new ToolStripSeparator(),
        new ToolStripMenuItem("終了(&X)", null, delegate { ExitThread(); }),
      });

      haltOperationWorker = new HaltOperationBackgroundWorker();
      haltOperationWorker.WorkerReportsProgress = true;
      haltOperationWorker.WorkerSupportsCancellation = true;
      haltOperationWorker.ProgressChanged += delegate(object sender, ProgressChangedEventArgs e) {
        var progress = e.UserState as HaltOperationProgress;

        if (progress.Status == HaltOperationStatus.Halting &&
            runningHaltOption.ReportMailToAddress != null && runningHaltOption.IsSuspendLikeMethod) {
          try {
            ReportMailSender.SendReport(runningHaltOption, null);
          }
          catch (System.Net.Mail.SmtpException ex) {
            Console.Error.WriteLine(ex);
          }
          catch (Exception ex) {
            Console.Error.WriteLine(ex);
          }
        }

        if (MainForm != null)
          (MainForm as MainDialog).DisplayProgress(progress);

        if (!toolStripMenuItemStatusSeparator.Visible)
          toolStripMenuItemStatusSeparator.Visible = true;
        if (!toolStripMenuItemStatus.Visible)
          toolStripMenuItemStatus.Visible = true;

        toolStripMenuItemStatus.Text = progress.StatusText;
        Text = progress.StatusText;

        if (progress.NotifyProgress) {
          Console.Error.WriteLine(progress.StatusText);

          if (!(progress.Status == HaltOperationStatus.Ready && runningHaltOption.HaltNow))
            ShowBalloonTip(5000, Application.ProductName, progress.StatusText, ToolTipIcon.Info);
        }
      };

      haltOperationWorker.RunWorkerCompleted += delegate(object sender, RunWorkerCompletedEventArgs e) {
        if (cancelAndQuitRequired) {
          InternalExitThread();
          return;
        }

        toolStripMenuItemStatusSeparator.Visible = false;
        toolStripMenuItemStatus.Visible = false;
        Text = Application.ProductName;

        if (MainDialog != null)
          MainDialog.SetHaltableAllignment();

        if (e.Cancelled)
          return;

        var exception = (e.Error == null) ? (e.Result as HaltOperationResult).Exception : e.Error;

        if (runningHaltOption.ReportMailToAddress != null && !runningHaltOption.IsSuspendLikeMethod) {
          try {
            ReportMailSender.SendReport(runningHaltOption, exception);
          }
          catch (System.Net.Mail.SmtpException ex) {
            Console.Error.WriteLine(ex);
          }
          catch (Exception ex) {
            Console.Error.WriteLine(ex);
          }
        }

        if (exception == null) {
          // succeeded
          if (!runningHaltOption.DontQuit)
            InternalExitThread();
        }
        else {
          // unexpected exception thrown or error occured while halting
          Console.Error.WriteLine(exception);

          if (runningHaltOption.ForceQuit)
            InternalExitThread();
        }
      };

      Text = Application.ProductName;

      if (option == null) {
        Console.Error.WriteLine("停止方法が設定されていないか、引数が不正なため、設定ダイアログを開きます。");

        ShowDialog();
      }
      else {
        StartHaltOperation(option);

        if (option.ShowWindow)
          ShowDialog();
      }
    }

    protected override void Dispose(bool disposing)
    {
      if (disposing) {
        if (components != null)
          components.Dispose();
        if (haltOperationWorker != null)
          haltOperationWorker.Dispose();
      }

      haltOperationWorker = null;

      base.Dispose(disposing);
    }

    protected override void ExitThreadCore()
    {
      cancelAndQuitRequired = false;

      if (haltOperationWorker.IsBusy) {
        var result = MessageBox.Show(MainDialog, string.Format("{0}を中断しますか?", runningHaltOption.MethodString), Application.ProductName, MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);

        if (result == DialogResult.Yes) {
          cancelAndQuitRequired = true;
          haltOperationWorker.CancelAsync();
        }

        return;
      }

      base.ExitThreadCore();
    }

    private void InternalExitThread()
    {
      if (MainForm != null)
        MainForm = null;

      ExitThread();
    }

    protected override void OnDoubleClick(EventArgs e)
    {
      ShowDialog();

      base.OnDoubleClick(e);
    }

#region "context menu"
    protected override void OnMouseClick(MouseEventArgs e)
    {
      ShowContextMenu(e.Button);

      base.OnMouseClick(e);
    }

    protected override ContextMenuStrip GetContextMenu(MouseButtons button)
    {
      if (button == MouseButtons.Left || button == MouseButtons.Right)
        return contextMenu;
      else
        return null;
    }
#endregion

    internal void AbortHaltOperation()
    {
      haltOperationWorker.CancelAsync();
    }

    internal void StartHaltOperation(HaltOption option)
    {
      runningHaltOption = option.Clone();

      // save current configurations
      Config.Save();

      haltOperationWorker.RunWorkerAsync(option);

      if (MainDialog != null)
        MainDialog.SetAbortableAllignment();
    }

    private void ShowDialog()
    {
      if (MainDialog == null) {
        MainForm = new MainDialog();
        MainDialog.AppContext = this;
        MainDialog.Show();
      }
      else {
        if (MainDialog.WindowState == FormWindowState.Minimized)
          MainDialog.WindowState = FormWindowState.Normal;
        MainDialog.Activate();
      }

      MainDialog.ApplyOption(runningHaltOption);
    }

    private Container components;
    private ContextMenuStrip contextMenu;
    private ToolStripMenuItem toolStripMenuItemStatus;
    private ToolStripSeparator toolStripMenuItemStatusSeparator;
    private HaltOperationBackgroundWorker haltOperationWorker;
    private HaltOption runningHaltOption = null;
    private bool cancelAndQuitRequired = false;
  }
}
